from service_locator_api.shared.error import CodedError, RegisterError
from messagebus import AckMessage, NakMessage, BadRequestMessage
from messageprocessorutil import UuidMessageProcessor
from .dto import RegisterServiceRequest

from .._ServiceLocator import ServiceLocator


class RegisterServiceProcessor(UuidMessageProcessor):

    def __init__(self, service_locator: ServiceLocator) -> None:
        super(RegisterServiceProcessor, self).__init__()
        self.service_locator = service_locator

    def parse_data(self, data):
        if data == b"":
            raise RegisterError.EMPTY_SERVICE_NAME

        parts = data.split(b"\0")
        if len(parts) < 2:
            raise RegisterError.EMPTY_ADDRESS

        try:
            service_name = parts[0].decode("utf-8")
        except UnicodeDecodeError:
            raise RegisterError.INVALID_SERVICE_NAME

        try:
            service_address = parts[1].decode("utf-8")
        except UnicodeDecodeError:
            raise RegisterError.ADDRESS_CANNOT_BE_ENCODED_USING_UTF8

        return RegisterServiceRequest(service_name, service_address)

    def call_business(self, request):
        # type: (RegisterServiceRequest) -> None
        return self.service_locator.register_service(request.service_name, request.address)

    def format_response(self, message_bus, message, input_obj, reg_id):
        message_bus.reply_message(message, AckMessage(reg_id.encode("utf-8")))

    def format_exception(self, message_bus, message, input_obj, exception):
        message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))

    def format_parse_exception(self, message_bus, message, data, exception):
        if isinstance(exception, CodedError):
            message_bus.reply_message(
                message=message,
                reply_message=BadRequestMessage(
                    error_code=exception.code,
                    error_message=exception.message,
                )
            )
        else:
            message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))



from messagebus import TokenCreator, TokenPriority

SERVICE_LOCATOR_GROUP = "3"
TK_SERVICE_LOCATOR_REGISTER_SERVICE = TokenCreator.create_token(TokenPriority.high, SERVICE_LOCATOR_GROUP, "1")
TK_SERVICE_LOCATOR_LOOKUP_SERVICE = TokenCreator.create_token(TokenPriority.high, SERVICE_LOCATOR_GROUP, "2")
TK_SERVICE_LOCATOR_KEEP_ALIVE = TokenCreator.create_token(TokenPriority.high, SERVICE_LOCATOR_GROUP, "3")



import logging
from threading import (
    Condition,
    Thread,
)
from typing import (
    Optional,
)

from ._Registration import Registration
from ._RegistrationRepository import RegistrationRepository

logger = logging.getLogger("ZeroMqServiceLocator.ServiceLocator")


class ServiceLocator(object):
    def __init__(
        self,
        registration_repository: RegistrationRepository,
        registration_expiration: float = 20,
    ):
        self.registration_repository = registration_repository
        self.clean_interval = registration_expiration / 2
        self.registration_expiration = registration_expiration

        self.clean_thread_running = True
        self.clean_thread_condition = Condition()
        self.clean_thread = Thread(target=self._run_clean_expired_registrations,
                                   name="ServiceLocatorCleanExpiredRegistrations")
        self.clean_thread.daemon = True
        self.clean_thread.start()

    def register_service(self, service_name, address):
        address = address.replace("*", "localhost")
        registration = Registration(
            service_name=service_name,
            service_address=address,
            registration_expiration=self.registration_expiration,
        )
        logger.debug("Created registration id: %s", registration.reg_id)

        self.registration_repository.add_registration(registration=registration)

        logger.debug("Current registrations: %s", self.registration_repository.get_all_registrations())
        logger.debug("Current services: %s", self.registration_repository.get_registered_services())

        return registration.reg_id

    def locate_service(self, service_name: str) -> Optional[str]:
        return self.registration_repository.lookup(service_name=service_name)

    def keep_alive(self, reg_id):
        self.registration_repository.keep_alive(reg_id=reg_id)

    def dispose(self):
        self.clean_thread_running = False
        with self.clean_thread_condition:
            self.clean_thread_condition.notify_all()
        self.clean_thread.join()

    def _run_clean_expired_registrations(self):
        while self.clean_thread_running:
            try:
                self.registration_repository.clean_registrations()
                if not self.clean_thread_running:
                    return

                with self.clean_thread_condition:
                    self.clean_thread_condition.wait(self.clean_interval)
            except (Exception,):
                logger.exception("Error cleaning expired registrations")



import logging
import os
import signal
from concurrent.futures.thread import ThreadPoolExecutor

import zmq
from helper import (
    config_logger,
    read_json_from_file_as_dictionary,
)
from messagehandlerbuilder import (
    Configuration,
    RedisConfig,
    ServiceLocatorMessageHandlerBuilder,
    StoreStrategy,
)
from messagereceiver import (
    MultiStrategyMessageReceiver,
    ThreadPoolExecutorHandleMessageStrategy,
)
from zeromqlowlevelmessagebus import ZeroMqLowLevelMessageBus

COMP_NAME = "ServiceLocator"

logger = logging.getLogger(__name__)


def main():
    config_path = './config.json'
    if "CONFIG_PATH" in os.environ:
        config_path = os.environ["CONFIG_PATH"]

    if os.path.isfile(config_path):
        config_dict = read_json_from_file_as_dictionary(config_path)
    else:
        config_dict = {}

    port = config_dict.get("port", 6000)
    store_strategy = StoreStrategy.IN_MEMORY
    redis_config = None
    if "redis" in config_dict:
        store_strategy = StoreStrategy.REDIS
        redis_dict = config_dict["redis"]
        redis_config = RedisConfig(
            host=redis_dict["host"],
            port=redis_dict["port"],
            db=redis_dict["db"],
            ssl=redis_dict["ssl"],
        )
    registration_expiration = 20
    if "registrationExpiration" in config_dict:
        registration_expiration = config_dict["registrationExpiration"]

    log_path = config_path
    if "LOG_PATH" in os.environ:
        log_path = os.environ["LOG_PATH"]

    config_logger(loader_path=log_path)

    context = zmq.Context()

    message_bus = ZeroMqLowLevelMessageBus(
        context=context,
        port=port,
    )

    message_handler_builder = ServiceLocatorMessageHandlerBuilder(
        config=Configuration(
            store_strategy=store_strategy,
            registration_expiration=registration_expiration,
            redis_config=redis_config,
        ),
    )

    message_receiver = MultiStrategyMessageReceiver(
        message_bus=message_bus,
        message_handler_builder=message_handler_builder,
        poll_timeout=1000,
        handle_strategy=ThreadPoolExecutorHandleMessageStrategy(thread_pool_executor=ThreadPoolExecutor(max_workers=50))
    )

    def signal_handler(signum, frame):
        print(f"Received {signum}")
        logger.info(f"Received {signum}")
        message_receiver.stop()

    signal.signal(signal.SIGINT, signal_handler)

    message_receiver.receive_messages()
    logger.info("Terminei")
    print("Terminou")


if __name__ == '__main__':
    main()



from ._InMemoryRegistrationRepository import InMemoryRegistrationRepository
from ._KeepAliveError import KeepAliveError
from ._RedisRegistrationRepository import RedisRegistrationRepository
from ._ServiceLocator import ServiceLocator



import logging
from typing import Optional

from messagehandler import MessageHandlerBuilder
from messageprocessor import (
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from messageprocessorutil import LoggingProcessorCallback
from service_locator_api.shared.token import (
    TK_SERVICE_LOCATOR_KEEP_ALIVE,
    TK_SERVICE_LOCATOR_LOOKUP_SERVICE,
    TK_SERVICE_LOCATOR_REGISTER_SERVICE,
)
from zeromqservicelocator import (
    InMemoryRegistrationRepository,
    RedisRegistrationRepository,
    ServiceLocator,
)
from zeromqservicelocator.processor import (
    KeepAliveProcessor,
    LocateServiceProcessor,
    RegisterServiceProcessor,
)
from .config import (
    Configuration,
    StoreStrategy,
)


class ServiceLocatorMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(
        self,
        config: Configuration,
    ) -> None:
        self.config = config

        self.service_locator: Optional[ServiceLocator] = None
        self.registration_repository: Optional[RedisRegistrationRepository] = None

    def build_singletons(self):
        if self.config.store_strategy == StoreStrategy.IN_MEMORY:
            self.registration_repository = InMemoryRegistrationRepository()
        else:
            self.registration_repository = RedisRegistrationRepository(
                host=self.config.redis_config.host,
                port=self.config.redis_config.port,
                db=self.config.redis_config.db,
                expiration=self.config.registration_expiration,
            )
        self.service_locator = ServiceLocator(
            registration_repository=self.registration_repository,
            registration_expiration=self.config.registration_expiration,
        )

    def build_message_handler(self, message):
        handled_tokens = {
            TK_SERVICE_LOCATOR_REGISTER_SERVICE: RegisterServiceProcessor(self.service_locator),
            TK_SERVICE_LOCATOR_LOOKUP_SERVICE: LocateServiceProcessor(self.service_locator),
            TK_SERVICE_LOCATOR_KEEP_ALIVE: KeepAliveProcessor(self.service_locator),
        }
        return MessageProcessorMessageHandler(
            message_processors=handled_tokens,
            executor_factory=DefaultMessageProcessorExecutorFactory(
                [
                    LoggingProcessorCallback(logging.getLogger("servicelocator.LoggingProcessorCallback"))
                ]
            )
        )

    def destroy_message_handler(self, message_handler):
        pass

    def destroy_singletons(self):
        self.service_locator.dispose()
        if self.config.store_strategy == StoreStrategy.REDIS:
            self.registration_repository.close()



from typing import Optional

from messagebus import (
    AckMessage,
    ApplicationErrorMessage,
    BadRequestMessage,
    NakMessage,
)
from messageprocessorutil import UuidMessageProcessor
from service_locator_api.shared.error import (
    CodedError,
    LocateError,
)
from .._ServiceLocator import ServiceLocator


class LocateServiceProcessor(UuidMessageProcessor):

    def __init__(self, service_locator: ServiceLocator) -> None:
        super(LocateServiceProcessor, self).__init__()
        self.service_locator = service_locator

    def parse_data(self, data):
        try:
            service_name = data.decode("utf-8")
        except UnicodeDecodeError:
            raise LocateError.INVALID_SERVICE_NAME

        return service_name

    def call_business(self, service_name: str) -> Optional[str]:
        return self.service_locator.locate_service(service_name)

    def format_response(self, message_bus, message, service_name, service_address):
        if service_address is not None:
            message_bus.reply_message(message, AckMessage(data=service_address.encode("utf-8")))
        else:
            message_bus.reply_message(
                message=message,
                reply_message=ApplicationErrorMessage(
                    error_code=LocateError.SERVICE_NOT_FOUND.code,
                    error_message=LocateError.SERVICE_NOT_FOUND.message,
                )
            )

    def format_exception(self, message_bus, message, input_obj, exception):
        message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))

    def format_parse_exception(self, message_bus, message, data, exception):
        if isinstance(exception, CodedError):
            message_bus.reply_message(
                message=message,
                reply_message=BadRequestMessage(
                    error_code=exception.code,
                    error_message=exception.message,
                )
            )
        else:
            message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))



import uuid
from datetime import (
    datetime,
    timedelta,
)


class Registration(object):
    def __init__(
        self,
        service_name: str,
        service_address: str,
        registration_expiration: float,
    ):
        self.reg_id = str(uuid.uuid4())
        self.service_name = service_name
        self.service_address = service_address
        self.registration_expiration = registration_expiration
        self.last_keep_alive = datetime.now()

    def keep_alive(self):
        self.last_keep_alive = datetime.now()

    def is_expired(self):
        return self.last_keep_alive + timedelta(seconds=self.registration_expiration) < datetime.now()

    def __str__(self):
        return str({
            "reg_id": self.reg_id,
            "service_name": self.service_name,
            "service_address": self.service_address,
            "registration_expiration": self.registration_expiration,
            "last_keep_alive": self.last_keep_alive.isoformat()
        })

    def __repr__(self) -> str:
        return self.__str__()



import logging
from threading import Lock
from typing import (
    Dict,
    Optional,
    Sequence,
)

from opentelemetry import trace
from opentelemetry.trace import Span
from service_locator_api.shared.exception import RegistrationNotFound
from ._KeepAliveError import KeepAliveError
from ._Registration import Registration
from ._RegistrationRepository import RegistrationRepository

logger = logging.getLogger("ZeroMqServiceLocator.InMemoryRegistrationRepository")
tracer = trace.get_tracer(__name__)


class InMemoryRegistrationRepository(RegistrationRepository):
    def __init__(self):
        self.services: Dict[str, RoundRobinList] = {}
        self.registrations: Dict[str, Registration] = {}

        self.lock = Lock()

    def add_registration(self, registration: Registration):
        with self.lock:
            if registration.service_name not in self.services:
                self.services[registration.service_name] = RoundRobinList()

            self.services[registration.service_name].append(registration.service_address)

            self.registrations[registration.reg_id] = registration

    def lookup(self, service_name: str) -> Optional[str]:
        with self.lock:
            if service_name not in self.services:
                return None

            address_to_return = self.services[service_name].get()

            return address_to_return

    def keep_alive(self, reg_id: str):
        try:
            with self.lock:
                registration_info = self.registrations[reg_id]
                registration_info.keep_alive()

            logger.debug("Keep alive processed: %s", registration_info)
            logger.debug("Current registrations: %s", self.registrations)
            logger.debug("Current services: %s", self.services)
        except KeyError as k:
            if k.args[0] == reg_id:
                raise RegistrationNotFound()

            logger.debug("Current registrations: %s", self.registrations)
            logger.debug("Current services: %s", self.services)
            raise KeepAliveError("The registration is active but the service is not registered.".format(reg_id))

    def clean_registrations(self):
        logger.debug("Running clean")
        logger.debug("Current registrations: %s", self.registrations)
        logger.debug("Current services: %s", self.services)
        with self.lock:
            with tracer.start_as_current_span("RegistrationsCleaned"):
                to_remove = []
                services_to_keep = {}
                for reg_id in self.registrations.keys():
                    registration_info = self.registrations[reg_id]
                    if registration_info.is_expired():
                        to_remove.append(registration_info)
                    else:
                        services_to_keep[registration_info.service_name] = registration_info.service_name

                for registration_info in to_remove:
                    logger.debug("Removing expired registration: %s", registration_info)

                    with tracer.start_as_current_span("RegistrationCleaned") as span:  # type: Span
                        span.set_attribute("registrationId", registration_info.reg_id)
                        del self.registrations[registration_info.reg_id]

                    self.services[registration_info.service_name].remove(registration_info.service_address)
                    if len(self.services[registration_info.service_name]) == 0:
                        del self.services[registration_info.service_name]

        logger.debug("Cleaned services: %s", self.services)
        logger.debug("Cleaned registrations: %s", self.registrations)

    def get_all_registrations(self) -> Sequence[Registration]:
        return tuple(list(self.registrations.values()))

    def get_registered_services(self) -> Sequence[str]:
        return tuple(self.services.keys())


class RoundRobinList(object):
    def __init__(self):
        self.index = -1
        self.data = []

    def append(self, data):
        self.data.append(data)

    def remove(self, data):
        self.data.remove(data)

    def get(self):
        self.index += 1
        if self.index >= len(self.data):
            self.index = 0

        return self.data[self.index]

    def __len__(self):
        return len(self.data)



from enum import Enum
from typing import Optional


class StoreStrategy(Enum):
    IN_MEMORY = "IN_MEMORY"
    REDIS = "REDIS"


class RedisConfig(object):
    def __init__(
        self,
        host: str,
        port: int,
        db: int,
        ssl: bool,
    ):
        self.host = host
        self.port = port
        self.db = db
        self.ssl = ssl


class Configuration(object):
    def __init__(
        self,
        store_strategy: StoreStrategy,
        registration_expiration: float,
        redis_config: Optional[RedisConfig] = None,
    ):
        if store_strategy == StoreStrategy.REDIS and redis_config is None:
            raise RuntimeError("When using Redis the Redis configuration must be sent")

        self.store_strategy = store_strategy
        self.registration_expiration = registration_expiration
        self.redis_config = redis_config



from keepalive_tests import *  # noqa: We want all tests to run
from locate_tests import *  # noqa: We want all tests to run
from messagehandlerbuilder import StoreStrategy
from register_tests import *  # noqa: We want all tests to run


@pytest.fixture
def store_strategy():
    return StoreStrategy.IN_MEMORY



from dockerredis import DockerRedis
from keepalive_tests import *  # noqa: We want all tests to run
from locate_tests import *  # noqa: We want all locate tests to run with the distributes configuration
from locate_tests import *  # noqa: We want all tests to run
from messagehandlerbuilder import (
    RedisConfig,
    StoreStrategy,
)
from redis import Redis
from register_tests import *  # noqa: We want all tests to run


@pytest.fixture(scope="module", autouse=True)
def redis():
    docker = DockerRedis()

    yield

    docker.stop()


@pytest.fixture(autouse=True)
def clean_redis(redis, redis_config):
    yield

    client = Redis(
        host=redis_config.host,
        port=redis_config.port,
        db=redis_config.db
    )
    client.flushdb()
    client.close()


@pytest.fixture
def store_strategy():
    return StoreStrategy.REDIS


@pytest.fixture()
def redis_config():
    return RedisConfig(
        host="localhost",
        port=6379,
        db=0,
        ssl=False,
    )



from typing import (
    Optional,
    Sequence,
)

from redis import Redis
from redis.client import Pipeline
from service_locator_api.shared.exception import RegistrationNotFound
from ._Registration import Registration
from ._RegistrationRepository import RegistrationRepository


class RedisRegistrationRepository(RegistrationRepository):
    registration_key_template = "serviceLocator:registration:{reg_id}"
    service_name_key_template = "serviceLocator:serviceName:{service_name}"
    all_services_key = "serviceLocator:allServices"

    def __init__(
        self,
        host: str,
        port: int,
        db: int,
        expiration: float
    ):
        self.client = Redis(
            host=host,
            port=port,
            db=db,
            decode_responses=True,
        )
        self.expiration = int(expiration * 1000)

    def add_registration(self, registration: Registration):
        registration_key = self.registration_key_template.format(reg_id=registration.reg_id)
        service_name_key = self.service_name_key_template.format(service_name=registration.service_name)

        def add_registration_in_transaction(pipeline: Pipeline):
            pipeline.multi()
            pipeline.set(
                name=registration_key,
                value=str(registration),
                px=self.expiration,
            )
            pipeline.rpush(
                service_name_key,
                f"{registration.service_address}\0{registration.reg_id}",
            )
            pipeline.sadd(
                self.all_services_key,
                registration.service_name,
            )
            pipeline.execute()

        self.client.transaction(
            func=add_registration_in_transaction,
            watches=[self.all_services_key, registration_key, service_name_key],
        )

    def lookup(self, service_name: str) -> Optional[str]:
        service_name_key = self.service_name_key_template.format(service_name=service_name)

        def get_current_address_in_transaction(pipeline: Pipeline):
            index = 0
            ret_address = None
            while True:
                left_value = self.client.lindex(
                    name=service_name_key,
                    index=index
                )
                if left_value is None:
                    break

                current_address, reg_id = left_value.split("\0")

                registration_key = self.registration_key_template.format(reg_id=reg_id)
                pipeline.watch(registration_key)

                registration_data = pipeline.get(name=registration_key)
                if registration_data is None:
                    index += 1
                    continue

                ret_address = current_address
                break

            pipeline.multi()
            pipeline.lpop(
                name=service_name_key,
                count=index + 1,
            )
            if left_value is not None:
                pipeline.rpush(
                    service_name_key,
                    left_value,
                )
            pipeline.execute()

            return ret_address

        return self.client.transaction(
            func=get_current_address_in_transaction,
            watches=[service_name_key],
            value_from_callable=True,
        )

    def keep_alive(self, reg_id: str):
        response = self.client.expire(
            name=self.registration_key_template.format(reg_id=reg_id),
            time=self.expiration,
        )
        if response == 0:
            raise RegistrationNotFound()

    def clean_registrations(self):
        pass

    def get_all_registrations(self) -> Sequence[Registration]:
        return []

    def get_registered_services(self) -> Sequence[str]:
        return []

    def close(self):
        self.client.close()



from service_locator_api.shared import error
from service_locator_api.shared.exception import (
    AddressChanged,
    RegistrationNotFound,
)
from service_locator_api.shared.token import TK_SERVICE_LOCATOR_ADDRESS_CHANGED
from messagebus import (
    AckMessage,
    Message,
    NakMessage,
    BadRequestMessage,
    ApplicationErrorMessage,
)
from messageprocessorutil import UuidMessageProcessor
from .. import KeepAliveError

from .._ServiceLocator import ServiceLocator


class KeepAliveProcessor(UuidMessageProcessor):

    def __init__(self, service_locator: ServiceLocator) -> None:
        super(KeepAliveProcessor, self).__init__()
        self.service_locator = service_locator

    def parse_data(self, data):
        try:
            return data.decode("utf-8")
        except UnicodeDecodeError:
            raise error.KeepAliveError.INVALID_REGISTRATION_ID

    def call_business(self, reg_id):
        # type: (str) -> None
        self.service_locator.keep_alive(reg_id)

    def format_response(self, message_bus, message, service_name, _):
        message_bus.reply_message(message, AckMessage())

    def format_exception(self, message_bus, message, input_obj, exception):
        if isinstance(exception, RegistrationNotFound):
            message_bus.reply_message(
                message=message,
                reply_message=ApplicationErrorMessage(
                    error_code=error.KeepAliveError.SUBSCRIPTION_ID_NOT_FOUND.code,
                    error_message=error.KeepAliveError.SUBSCRIPTION_ID_NOT_FOUND.message,
                )
            )
        elif isinstance(exception, AddressChanged):
            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=TK_SERVICE_LOCATOR_ADDRESS_CHANGED,
                    data=exception.current_address.encode(),
                )
            )
        elif isinstance(exception, KeepAliveError):
            message_bus.reply_message(message, NakMessage(exception.message.encode("utf-8")))
        else:
            message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))

    def format_parse_exception(self, message_bus, message, data, exception):
        if isinstance(exception, error.CodedError):
            message_bus.reply_message(
                message=message,
                reply_message=BadRequestMessage(
                    error_code=exception.code,
                    error_message=exception.message,
                )
            )
        else:
            message_bus.reply_message(message, NakMessage(str(exception).encode("utf-8")))



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Optional,
    Sequence,
)

from ._Registration import Registration


class RegistrationRepository(ABC):
    @abstractmethod
    def add_registration(self, registration: Registration):
        raise NotImplementedError()

    @abstractmethod
    def lookup(self, service_name: str) -> Optional[str]:
        raise NotImplementedError()

    @abstractmethod
    def keep_alive(self, reg_id: str):
        raise NotImplementedError()

    @abstractmethod
    def clean_registrations(self):
        raise NotImplementedError()

    @abstractmethod
    def get_all_registrations(self) -> Sequence[Registration]:
        raise NotImplementedError()

    @abstractmethod
    def get_registered_services(self) -> Sequence[str]:
        raise NotImplementedError()
